package kudos.ide.buffer;

import java.io.*;
import java.util.*;
import javax.swing.undo.*;
import javax.swing.text.*;
import javax.swing.event.*;

import kudos.ide.io.*;
import kudos.ide.form.*;
import kudos.ide.form.textarea.*;
import kudos.ide.syntax.*;
import kudos.ide.tools.*;

public class Buffer extends SyntaxDocument {
  private static final int TabSize = 4;

  private int flags;
  private int compoundEditCount;
  private long modTime;

  private static final int CLOSED = 0;
  private static final int LOADING = 1;
  private static final int SAVING = 2;
  private static final int NEW_FILE = 3;
  private static final int UNTITLED = 4;
  private static final int DIRTY = 6;
  private static final int READ_ONLY = 7;
  private static final int SYNTAX = 8;
  private static final int UNDO_IN_PROGRESS = 9;
  private static final int TEMPORARY = 10;

  public static final int IOBUFSIZE = 32768;

  private File file = null;
  private File autosaveFile = null;
  private String path = null;
  private String name = null;
  private String protocol = null;
  private CompoundEdit compoundEdit = null;
  private UndoManager undo = null;

  public static final String LINESEP = "lineSeparator";
  public static final String SELECTION_START = "Buffer__selStart";
  public static final String SELECTION_END = "Buffer__selEnd";

  private MainView mainView = null;
  private VirtualFileSystem vfs = null;

  public Buffer(MainView mainView, String path, boolean readOnly, 
		boolean newFile, boolean temp, Hashtable props)
  {
    this.mainView = mainView;
    undo = new UndoManager();
    setFlag(TEMPORARY,temp);
    setFlag(READ_ONLY,readOnly);

    addDocumentListener(new DocumentHandler());
    addUndoableEditListener(new UndoHandler());

    setPath(path);

    setFlag(UNTITLED,newFile);
    setFlag(NEW_FILE,newFile);
  }

  private void setPath(String path) {
    this.path = path;
    name = Util.getFileName(path);

    vfs = VirtualFileSystemManager.getVFSForProtocol(protocol);
  }
			
  public final String getPath() {
    return path;
  }

  public final String getName() {
    return name;
  }

  public final File getFile() {
    return file;
  }

  public boolean saveAs(MainView mainView) {
    String file = 
      VirtualFileSystemManager.getEditFileDialog().showSaveDialog(mainView,
								  this);
    if (file != null) {
      return save(mainView,file);
    }
    else {
      return false;
    }
  }

  public boolean load(final MainView mainView, final boolean reload) {
    setFlag(LOADING,true);

    undo = null;

    if (!getFlag(NEW_FILE)) {
      if (file != null) {
	modTime = file.lastModified();
      }

      if (!vfs.load(mainView,this,path)) {
	setFlag(LOADING,false);
	return false;
      }
    }

    VirtualFileSystemManager.runInAWTThread(new Runnable()
      {
	public void run()
	{
	  undo = new UndoManager();
	  setFlag(LOADING,false);
	}
      });

    return true;
  }

  public boolean save(final MainView mainView, String path) {
    if (path == null && getFlag(NEW_FILE)) {
      return saveAs(mainView);
    }

    setFlag(SAVING,true);

    if (path == null) {
      path = this.path;
    }

    if (!vfs.save(mainView,this,path)) {
      setFlag(SAVING,false);
      return false;
    }

    VirtualFileSystemManager.runInAWTThread(new Runnable() {
	public void run() {
	  setFlag(READ_ONLY,false);
	  setFlag(NEW_FILE,false);
	  setFlag(UNTITLED,false);
	  setFlag(DIRTY,false);
	  setFlag(SAVING,false);
	}
      });
    
    return true;
  }
  
  private void setFlag(int flag, boolean value) {
    if(value) {
      flags |= (1 << flag);
    }
    else {
      flags &= ~(1 << flag);
    }
  }

  public final boolean isNewFile() {
    return getFlag(NEW_FILE);
  }

  public final boolean isClosed() {
    return getFlag(CLOSED);
  }

  public final boolean isLoaded() {
    return !getFlag(LOADING);
  }

  public final boolean isDirty() {
    return getFlag(DIRTY);
  }

  public void setDirty(boolean d) {
    boolean old_d = getFlag(DIRTY);

    if (d) {
      if (getFlag(LOADING) || getFlag(READ_ONLY))
	return;
      setFlag(DIRTY,true);
    }
    else
      setFlag(DIRTY,false);
  }

  private boolean getFlag(int flag) {
    int mask = (1 << flag);
    return (flags & mask) == mask;
  }

  public void beginCompoundEdit() {
    if (getFlag(TEMPORARY)) {
      return;
    }

    compoundEditCount++;
    if(compoundEdit == null) {
      compoundEdit = new CompoundEdit();
    }
  }

  public void endCompoundEdit() {
    if (getFlag(TEMPORARY)) {
      return;
    }

    if (compoundEditCount == 0) {
      return;
    }
    compoundEditCount--;
    if (compoundEdit != null) {
      compoundEdit.end();
      if (compoundEdit.canUndo()) {
	undo.addEdit(compoundEdit);
      }
      compoundEdit = null;
    }
  }

  public String toString() {
    return path;
  }

  public int getTabSize() {
    return TabSize;
  }

  public boolean undo() {
    if(undo == null) {
      return false;
    }

    try {
      setFlag(UNDO_IN_PROGRESS,true);
      undo.undo();
      return true;
    } catch (CannotUndoException cu) {
      return false;
    } finally {
      setFlag(UNDO_IN_PROGRESS,false);
    }
  }

  public boolean redo() {
    if(undo == null) {
      return false;
    }

    try {
      setFlag(UNDO_IN_PROGRESS,true);
      undo.redo();
      return true;
    } catch (CannotRedoException cr) {
      return false;
    } finally {
      setFlag(UNDO_IN_PROGRESS,false);
    }
  }

  public Buffer _read(InputStream _in) 
    throws IOException, BadLocationException
  {
    int bufLength;
    if (file != null) {
      bufLength = (int)file.length();
    }
    else
      bufLength = IOBUFSIZE * 4;

    StringBuffer sbuf = new StringBuffer(bufLength);
    InputStreamReader in = new InputStreamReader(_in,"MS949");
    char[] buf = new char[IOBUFSIZE];
    int len;

    boolean CRLF = false;
    boolean CROnly = false;
    boolean lastWasCR = false;
		
    while ((len = in.read(buf,0,buf.length)) != -1) {
      int lastLine = 0;

      for (int i = 0; i < len; i++) {
	switch(buf[i]) {
	  case '\r':
	    if (lastWasCR) {
	      CROnly = true;
	      CRLF = false;
	    }
	    else {
	      lastWasCR = true;
	    }

	    sbuf.append(buf, lastLine, i-lastLine);
	    sbuf.append('\n');

	    lastLine = i + 1;
	    break;
	  case '\n':
	    if (lastWasCR) {
	      CROnly = false;
	      CRLF = true;
	      lastWasCR = false;
	      lastLine = i + 1;
	    }
	    else {
	      CROnly = false;
	      CRLF = false;
	      sbuf.append(buf,lastLine,
			  i - lastLine);
	      sbuf.append('\n');
	      lastLine = i + 1;
	    }
	    break;
	  default:
	    if (lastWasCR) {
	      CROnly = true;
	      CRLF = false;
	      lastWasCR = false;
	    }
	    break;
	}
      }
      sbuf.append(buf,lastLine,len - lastLine);
    }
		
    remove(0,getLength());
    in.close();
    String text = sbuf.toString();

    int start;
    MainControl mainControl = mainView.getMainControl();
    Buffer oldBuffer = mainControl.getOldBuffer();
    if (oldBuffer != null) {
      int caret = mainView.getEditTextArea().getCaretPosition();
      insertString(0,oldBuffer.getText(0,oldBuffer.getLength()),null);
      start = caret;
      mainControl.setOldBuffer(null);
      mainView.getEditTextArea().setDocumentCaret(start + text.length());
    } 
    else {
      start = 0;
    }
    insertString(start,text,null);
    setFlag(NEW_FILE,false);
    return this;
  }

  public void _write(OutputStream _out)
    throws IOException, BadLocationException 
  {
    BufferedWriter out = new BufferedWriter(
      new OutputStreamWriter(_out,"MS949"),IOBUFSIZE);
    Segment lineSegment = new Segment();
    String newline = "\r\n";
    Element map = getDefaultRootElement();
    for(int i = 0; i < map.getElementCount(); i++)
    {
      Element line = map.getElement(i);
      int start = line.getStartOffset();
      getText(start,line.getEndOffset() - start - 1,lineSegment);
      out.write(lineSegment.array,lineSegment.offset,lineSegment.count);
      out.write(newline);
    }
    out.close();
  }

  public void addUndoableEdit(UndoableEdit edit)
  {
    if(undo == null || getFlag(UNDO_IN_PROGRESS) || getFlag(LOADING))
      return;

    if(undo.canRedo() && !edit.isSignificant())
      return;

    if(compoundEdit != null)
      compoundEdit.addEdit(edit);
    else
      undo.addEdit(edit);
  }

  public boolean indentLine(EditTextArea editTextArea, int lineIndex, boolean force)
  {
    if(lineIndex == 0)
      return false;

    String openBrackets = "";
    String closeBrackets = "";
    String _indentPrevLine = null;
    boolean doubleBracketIndent = false;

    int tabSize = getTabSize();
    boolean noTabs = false;

    String line = editTextArea.getLineText(lineIndex);
    int start = editTextArea.getLineStartOffset(lineIndex);
    String prevLine = null;
    for(int i = lineIndex - 1; i >= 0; i--)
    {
      if(editTextArea.getLineLength(i) != 0)
      {
	prevLine = editTextArea.getLineText(i);
	break;
      }
    }

    if(prevLine == null)
      return false;

    boolean prevLineMatches = false;

    boolean prevLineStart = true;
    int prevLineIndent = 0;
    int prevLineBrackets = 0;
    for(int i = 0; i < prevLine.length(); i++)
    {
      char c = prevLine.charAt(i);
      switch(c)
      {
      case ' ':
	if(prevLineStart)
	  prevLineIndent++;
	break;
      case '\t':
	if(prevLineStart)
	{
	  prevLineIndent += (tabSize
			     - (prevLineIndent
				% tabSize));
	}
	break;
      default:
	prevLineStart = false;
	if(closeBrackets.indexOf(c) != -1)
	  prevLineBrackets = Math.max(
	    prevLineBrackets-1,0);
	else if(openBrackets.indexOf(c) != -1)
	{
	  if(!doubleBracketIndent)
	    prevLineMatches = false;
	  prevLineBrackets++;
	}
	break;
      }
    }

    boolean lineStart = true;
    int lineIndent = 0;
    int lineWidth = 0;
    int lineBrackets = 0;
    int closeBracketIndex = -1;

    for (int i = 0; i < line.length(); i++) {
      char c = line.charAt(i);
      switch(c) {
        case ' ':
	  if (lineStart) {
	    lineIndent++;
	    lineWidth++;
	  }
	  break;
        case '\t':
	  if (lineStart) {
	    lineIndent += (tabSize
			   - (lineIndent
			      % tabSize));
	    lineWidth++;
	  }
	  break;
        default:
	  lineStart = false;
	  if (closeBrackets.indexOf(c) != -1) {
	    if(lineBrackets == 0) {
	      closeBracketIndex = i;
	    }
	    else {
	    lineBrackets--;
	    }
	  }
	  else if(openBrackets.indexOf(c) != -1) {
	    if (!doubleBracketIndent) {
	      prevLineMatches = false;
	    }
	    lineBrackets++;
	  }
	  break;
      }
    }

    try {
      if (closeBracketIndex != -1) {
	int offset = TextUtilities.findMatchingBracket(
	  this,start + closeBracketIndex);
	if (offset != -1) {
	  String closeLine = editTextArea.getLineText(
	    editTextArea.getLineOfOffset(offset));
	  prevLineIndent = Util.getLeadingWhiteSpaceWidth(
	    closeLine,tabSize);
	}
	else {
	  return false;
	}
      }
      else {
	prevLineIndent += (prevLineBrackets * tabSize);
      }

      if (prevLineMatches) {
	prevLineIndent += tabSize;
      }

      if (!force && lineIndent >= prevLineIndent) {
	return false;
      }

      remove(start,lineWidth);
      insertString(start,Util.createWhiteSpace(
	prevLineIndent,(noTabs ? 0 : tabSize)),null);
      return true;
    } catch (BadLocationException bl) {
    }

    return false;
  }

  class UndoHandler implements UndoableEditListener
  {
    public void undoableEditHappened(UndoableEditEvent evt) {
      addUndoableEdit(evt.getEdit());
    }
  }

  class DocumentHandler implements DocumentListener
  {
    public void insertUpdate(DocumentEvent evt) {
      setDirty(true);
    }

    public void removeUpdate(DocumentEvent evt) {
      setDirty(true);
    }

    public void changedUpdate(DocumentEvent evt) {
    }
  }
}
