package kudos.ide.form.textarea;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;

public class LineNumber extends JComponent implements SwingConstants
{
  private int alignment;
  private int ileft = 0;
  private int baseline = 0;
  private int interval = 0;
  private boolean collapsed = false;
  private boolean lineNumberingEnabled = true;
  private boolean currentLineHighlightEnabled = false;

  private FontMetrics fm = null;
  private JPopupMenu context = null;
  private Dimension lineNumberSize = new Dimension(0,0);
  private Dimension collapsedSize = new Dimension(0,0);

  private EditTextArea editTextArea;

  public LineNumber(EditTextArea editTextArea)
  {
    this.editTextArea = editTextArea;

    setDoubleBuffered(true);

    GutterMouseListener ml = new GutterMouseListener();
    addMouseListener(ml);
    addMouseMotionListener(ml);
  }

  public void paintComponent(Graphics gfx)
  {
    if (!collapsed)
    {
      Rectangle r = gfx.getClipBounds();
      gfx.setColor(getBackground());
      gfx.fillRect(r.x, r.y, r.width, r.height);

      if (!editTextArea.getDocument().isLoaded()) return;

      if (lineNumberingEnabled) paintLineNumbers(gfx);
    }
  }

  private void paintLineNumbers(Graphics gfx)
  {
    FontMetrics pfm = editTextArea.getTextAreaPainter().getFontMetrics();
    int lineHeight = pfm.getHeight();

    Rectangle clip = gfx.getClipBounds();

    int baseline = (clip.y - clip.y % lineHeight) + (int) Math.round(
      (this.baseline + lineHeight - pfm.getDescent()) / 2.0);

    int firstLine = clip.y / lineHeight + editTextArea.getFirstLine() + 1;
    int lastLine = firstLine + clip.height / lineHeight;
    int caretLine = editTextArea.getCaretLine() + 1;

    int firstValidLine = firstLine > 1 ? firstLine : 1;
    int lastValidLine = (lastLine > editTextArea.getLineCount())
      ? editTextArea.getLineCount() : lastLine;


    gfx.setFont(getFont());

    String number;
    int offset;

    for (int line = firstLine; line <= lastLine;
	 line++, baseline += lineHeight)
    {
      if (line < firstValidLine || line > lastValidLine)
	continue;

      number = Integer.toString(line);

      offset = lineNumberSize.width - collapsedSize.width
	- (fm.stringWidth(number) + 1);

      gfx.setColor(Color.black);

      gfx.drawString(number, ileft + offset, baseline);
    }
  }

  public void setBorder(int width, Color color1, Color color2)
  {
    setBorder(new CompoundBorder(new MatteBorder(0,0,0,width,color2),
				 new MatteBorder(0,0,0,width,color1)));
  }

  public void setBorder(Border border)
  {
    super.setBorder(border);

    if (border == null)
    {
      ileft = 0;
      collapsedSize.width = 0;
      collapsedSize.height = 0;
    }
    else
    {
      Insets insets = border.getBorderInsets(this);
      ileft = insets.left;
      collapsedSize.width = insets.left + insets.right;
      collapsedSize.height = insets.top + insets.bottom;
    }
  }

  public boolean isCollapsed()
  {
    return collapsed;
  }

  public boolean isCurrentLineHighlightEnabled()
  {
    return currentLineHighlightEnabled;
  }

  public void setLineNumberWidth(int width)
  {
    if (width < collapsedSize.width) width = collapsedSize.width;

    lineNumberSize.width = width;

    if (!collapsed) editTextArea.revalidate();
  }

  public void toggleCollapsed()
  {
    setCollapsed(!collapsed);
  }

  public void setLineNumberingEnabled(boolean enabled)
  {
    if (lineNumberingEnabled == enabled) return;

    lineNumberingEnabled = enabled;

    repaint();
  }

  public void setCollapsed(boolean collapsed)
  {
    if (this.collapsed == collapsed) return;

    this.collapsed = collapsed;

    editTextArea.revalidate();
  }

  public final void invalidateLineRange(int firstLine, int lastLine)
  {
    FontMetrics pfm = editTextArea.getTextAreaPainter().getFontMetrics();
    repaint(0,editTextArea.lineToY(firstLine) + pfm.getDescent() + pfm.getLeading(),
	    getWidth(),(lastLine - firstLine + 1) * pfm.getHeight());
  }

  public final void invalidateLine(int line)
  {
    FontMetrics pfm = editTextArea.getTextAreaPainter().getFontMetrics();
    repaint(0,editTextArea.lineToY(line) + pfm.getDescent() + pfm.getLeading(),
	    getWidth(),pfm.getHeight());
  }

  class GutterMouseListener extends MouseAdapter
    implements MouseMotionListener
  {
    public void mouseClicked(MouseEvent e)
    {
      int count = e.getClickCount();
      if (count == 1)
      {
	if (context == null || context.isVisible())
	  return;

	if ((e.getModifiers() & InputEvent.BUTTON3_MASK)
	    != 0)
	{
	  editTextArea.requestFocus();

	  context.show(LineNumber.this,
		       e.getX(), e.getY());
	}
      }
      else if (count >= 2)
      {
	toggleCollapsed();
      }
    }

    public void mousePressed(MouseEvent e)
    {
      dragStart = e.getPoint();

      startWidth = lineNumberSize.width;
    }

    public void mouseDragged(MouseEvent e)
    {
      if (dragStart == null) return;

      if (isCollapsed()) setCollapsed(false);

      Point p = e.getPoint();

      lineNumberSize.width = startWidth + p.x - dragStart.x;

      if (lineNumberSize.width < collapsedSize.width)
      {
	lineNumberSize.width = startWidth;
	setCollapsed(true);
      }

      SwingUtilities.invokeLater(new Runnable() {
	  public void run()
	  {
	    editTextArea.revalidate();
	  }
	});
    }

    public void mouseExited(MouseEvent e)
    {
      if (dragStart != null && dragStart.x > e.getPoint().x)
      {
	setCollapsed(true);
	lineNumberSize.width = startWidth;

	SwingUtilities.invokeLater(new Runnable() {
	    public void run()
	    {
	      editTextArea.revalidate();
	    }
	  });
      }

    }

    public void mouseMoved(MouseEvent e) {}

    public void mouseReleased(MouseEvent e)
    {
      dragStart = null;
    }

    private Point dragStart = null;
    private int startWidth = 0;
  }
}
