package kudos.ide.form.textarea;

import java.awt.*;
import javax.swing.*;
import javax.swing.text.*;

import kudos.ide.syntax.*;

public class TextAreaPainter extends JComponent implements TabExpander
{
	private boolean blockCaret;
	private boolean eolMarkers;
	private boolean lineHighlight;
	private int tabSize;

	private FontMetrics fm = null;
	private Color selectionColor = new Color(200,200,255);
	private Color lineHighlightColor = new Color(240,240,255);
	
	private EditTextArea editTextArea = null;

	public TextAreaPainter(EditTextArea editTextArea)
	{
		this.editTextArea = editTextArea;

		setAutoscrolls(true);

		setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));

		setFont(new Font("Monospaced",Font.PLAIN,14));
		setBackground(Color.white);
	}

	public float nextTabStop(float x, int tabOffset)
	{
		int offset = editTextArea.getHorizontalOffset();
		int ntabs = ((int)x - offset) / tabSize;
		return (ntabs + 1) * tabSize + offset;
	}

	public void setFont(Font font)
	{
		super.setFont(font);
		fm = Toolkit.getDefaultToolkit().getFontMetrics(font);
	}

	public FontMetrics getFontMetrics()
	{
		return fm;
	}

	private void paintHighlight(Graphics gfx, int line, int y, boolean invalid)
	{
		if(!invalid && line >= editTextArea.getSelectionStartLine()
			&& line <= editTextArea.getSelectionEndLine())
			paintLineHighlight(gfx,line,y);

		if(!invalid && line == editTextArea.getCaretLine())
			paintCaret(gfx,line,y);
	}

	private void paintLineHighlight(Graphics gfx, int line, int y)
	{
		int height = fm.getHeight();
		y += fm.getLeading() + fm.getDescent();

		int selectionStart = editTextArea.getSelectionStart();
		int selectionEnd = editTextArea.getSelectionEnd();

		if(selectionStart == selectionEnd)
		{
			gfx.setColor(lineHighlightColor);
			gfx.fillRect(0,y,getWidth(),height);
		}
		else
		{
			gfx.setColor(selectionColor);

			int selectionStartLine = editTextArea.getSelectionStartLine();
			int selectionEndLine = editTextArea.getSelectionEndLine();
			int lineStart = editTextArea.getLineStartOffset(line);

			int x1, x2;
			if(editTextArea.isSelectionRectangular())
			{
				int lineLen = editTextArea.getLineLength(line);
				x1 = editTextArea.offsetToX(line,Math.min(lineLen,
					selectionStart - editTextArea.getLineStartOffset(
					selectionStartLine)));
				x2 = editTextArea.offsetToX(line,Math.min(lineLen,
					selectionEnd - editTextArea.getLineStartOffset(
					selectionEndLine)));
				if(x1 == x2)
					x2++;
			}
			if(selectionStartLine == selectionEndLine)
			{
				x1 = editTextArea.offsetToX(line, selectionStart - lineStart);
				x2 = editTextArea.offsetToX(line, selectionEnd - lineStart);
			}
			else if(line == selectionStartLine)
			{
				x1 = editTextArea.offsetToX(line, selectionStart - lineStart);
				x2 = getWidth();
			}
			else if(line == selectionEndLine)
			{
				x1 = 0;
				x2 = editTextArea.offsetToX(line, selectionEnd - lineStart);
			}
			else
			{
				x1 = 0;
				x2 = getWidth();
			}

			gfx.fillRect(x1 > x2 ? x2 : x1,y,x1 > x2 ?
				(x1 - x2) : (x2 - x1),height);
		}
	}

	public final boolean isBlockCaretEnabled()
	{
		return blockCaret;
	}

	private void paintLine(Graphics gfx, int line, int x)
	{
		Font defaultFont = getFont();
		Color defaultColor = getForeground();

		int y = editTextArea.lineToY(line);

		boolean invalid = (!editTextArea.getDocument().isLoaded() || line < 0
			|| line >= editTextArea.getLineCount());
		paintHighlight(gfx,line,y,invalid);

		paintPlainLine(gfx,line,defaultFont,defaultColor,x,y);
	}

	private void paintPlainLine(Graphics gfx, int line, Font defaultFont,
		Color defaultColor, int x, int y)
	{
		gfx.setFont(defaultFont);
		gfx.setColor(defaultColor);

		y += fm.getHeight();
		editTextArea.getLineText(line,editTextArea.lineSegment);
		x = Utilities.drawTabbedText(editTextArea.lineSegment,x,y,gfx,this,0);

		if(eolMarkers)
		{
			gfx.setColor(Color.black);
			gfx.drawString(".",x,y);
		}
	}

	private void paintCaret(Graphics gfx, int line, int y)
	{
		if(editTextArea.isCaretVisible())
		{
			int offset = editTextArea.getCaretPosition() - editTextArea.getLineStartOffset(line);
			int caretX = editTextArea.offsetToX(line,offset);
			int caretWidth = ((blockCaret || editTextArea.isOverwriteEnabled()) ?
				fm.charWidth('w') : 1);
			y += fm.getLeading() + fm.getDescent();
			int height = fm.getHeight();
			
			gfx.setColor(Color.black);
			gfx.drawRect(caretX,y,caretWidth,height - 1);
		}
	}

	public void paint(Graphics gfx)
	{
		tabSize = fm.charWidth(' ') * ((Integer)editTextArea
			.getDocument().getProperty(
			PlainDocument.tabSizeAttribute)).intValue();

		Rectangle clipRect = gfx.getClipBounds();

		gfx.setColor(getBackground());
		gfx.fillRect(clipRect.x,clipRect.y,clipRect.width,clipRect.height);

		int height = fm.getHeight();
		int firstLine = editTextArea.getFirstLine();
		int firstInvalid = firstLine + clipRect.y / height;
		int lastInvalid = firstLine + (clipRect.y + clipRect.height - 1) / height;

		try
		{
			int x = editTextArea.getHorizontalOffset();

			for(int line = firstInvalid; line <= lastInvalid; line++)
			{
				paintLine(gfx,line,x);
			}
		}
		catch(Exception e) {}
	}

	public final void invalidateLine(int line)
	{
		repaint(0,editTextArea.lineToY(line) + fm.getDescent() + fm.getLeading(),
			getWidth(),fm.getHeight());
	}

	public final void invalidateSelectedLines()
	{
		invalidateLineRange(editTextArea.getSelectionStartLine(),
			editTextArea.getSelectionEndLine());
	}

	public final void invalidateLineRange(int firstLine, int lastLine)
	{
		repaint(0,editTextArea.lineToY(firstLine) + fm.getDescent() + fm.getLeading(),
			getWidth(),(lastLine - firstLine + 1) * fm.getHeight());
	}
}