/* -*---------------------------------------------------------
 * Uses: Classic VM (build JDK-1.2-V, green threads, sunwjit)
 * -*---------------------------------------------------------
 * @(#) $Id$
 *
 * Evaluator.java:
 *   - created by Cheoljoo Jeong
 *
 * Copyleft(c) 2000. Cheoljoo Jeong. All lefts reserved.
 *
 * $Log$
 */

package kudos.schemej;

import kudos.schemej.util.*;
import kudos.lang.*;
import kudos.jvm.*;
import kudos.exns.*;

/**
 * An evaluator for SchemeJ expresions. Evaluates the expression and
 * returns the value, an instance of the Object class.
 * Note that Evaluator is a Singleton class.
 *
 * @author Cheoljoo Jeong
 */

public class Evaluator implements ExprVisitorI {
  public static Evaluator evaluator = new Evaluator();

  /**
   * constructor
   */
  public Evaluator() {
  }

  /**
   * Returns the single instance of this class.
   */
  public static Evaluator getInstance() {
    if (evaluator == null) {
      evaluator = new Evaluator();
    }
    return evaluator;
  }

  public Object visitExpr(Expr expr, Environment r, Continuation k)
    throws SchemeJException 
  {
    try {
      Object value = null;
      switch (expr.getExprType()) {
      case ExprType.TApplyExpr:
	value = visitApplyExpr((ApplyExpr) expr, r, k);
	break;
      case ExprType.TBeginExpr:
	value = visitBeginExpr((BeginExpr) expr, r, k);
	break;
      case ExprType.TCaseExpr:
	value = visitCaseExpr((CaseExpr) expr, r, k);
	break;
      case ExprType.TCondExpr:
	value = visitCondExpr((CondExpr) expr, r, k);
	break;
      case ExprType.TDelayExpr:
	value = visitDelayExpr((DelayExpr) expr, r, k);
	break;
      case ExprType.TDoExpr:
	value = visitDoExpr((DoExpr) expr, r, k);
	break;
      case ExprType.TIfExpr:
	value = visitIfExpr((IfExpr) expr, r, k);
	break;
      case ExprType.TLambdaExpr:
	value = visitLambdaExpr((LambdaExpr) expr, r, k);
	break;
      case ExprType.TLetExpr:
	value = visitLetExpr((LetExpr) expr, r, k);
	break;
      case ExprType.TLetStarExpr:
	value = visitLetStarExpr((LetStarExpr) expr, r, k);
	break;
      case ExprType.TLetrecExpr:
	value = visitLetrecExpr((LetrecExpr) expr, r, k);
	break;
      case ExprType.TLiteralExpr:
	value = visitLiteralExpr((LiteralExpr) expr, r, k);
	break;
      case ExprType.TQuasyQuoteExpr:
	value = visitQuasyQuoteExpr((QuasyQuoteExpr) expr, r, k);
	break;
      case ExprType.TQuoteExpr:
	value = visitQuoteExpr((QuoteExpr) expr, r, k);
	break;
      case ExprType.TSetExpr:
	value = visitSetExpr((SetExpr) expr, r, k);
	break;
      case ExprType.TVarRefExpr:
	value = visitVarRefExpr((VarRefExpr) expr, r, k);
	break;
      case ExprType.TDefineForm:
	value = visitDefineForm((DefineForm) expr, r, k);
	break;
      default:
	throw new SchemeJException(SchemeJException.SysError,
				   "unknown expression type");
      }
      return value;
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  
  }

  /**
   * Processes VarRefExpr class. Looks up the environment for
   * the binding which contains this variable and returns the 
   * corresponding value.
   */
  private Object visitVarRefExpr(VarRefExpr varRefExpr, Environment r,
				 Continuation k) 
    throws UnboundVarException, SchemeJException
  {
    try {
      // CLARIFY!!!
      Object value = r.lookup(varRefExpr.name);
      return k.resume(value);
    } catch (UnboundVarException e) {
      return k.resume(e);
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  }

  /**
   * Processes literal expressions. 
   */
  private Object visitLiteralExpr(LiteralExpr literalExpr, Environment r,
				  Continuation k) 
  {
    return k.resume(literalExpr.value);
  }

  /**
   * Processes ApplyExpr class. Applies the procedure to the set
   * of arguments given.
   */
  private Object visitApplyExpr(ApplyExpr applyExpr, Environment r,
				Continuation k) 
    throws InvalidRatorException, WrongNumberOfRandsException, 
	   SchemeJException
  {
    try {
      ApplyContinuation applyk = 
	new ApplyContinuation(applyExpr.rands, r, k);
      return applyExpr.rator.eval(this, r, applyk);
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  }

  /**
   * Processes quoted expressions. As seen in [Quiennec], the semantics
   * of a quoted expression is an SExpr.
   */
  private Object visitQuoteExpr(QuoteExpr quoteExpr, Environment r,
				Continuation k) 
  {
    return k.resume(quoteExpr.sexpr);
  }

  /**
   * Processes conditional expressions.
   */
  private Object visitIfExpr(IfExpr ifExpr, Environment r, Continuation k) 
    throws InvalidTypeException, SchemeJException
  {
    try {
      IfContinuation ifk = 
	new IfContinuation(ifExpr.thenExpr, ifExpr.elseExpr, r, k);
      return ifExpr.condExpr.eval(this, r, ifk);
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  }

  /**
   * Processes set! expressions.
   */
  private Object visitSetExpr(SetExpr setExpr, Environment r, Continuation k) 
    throws InvalidLvalException, SchemeJException
  {
    try {
      String name = setExpr.var.name;
      Continuation errork = new PrintErrorContinuation(r, null);
      r.update(setExpr.var.name,
	       setExpr.expr.eval(this, r, errork));
      return k.resume(setExpr.var);
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  }
  
  /**
   * Processes lambda abstractions.
   */
  private Object visitLambdaExpr(LambdaExpr lambdaExpr, Environment r,
				 Continuation k) 
  {
    // make closure from formals, body, and current environment
    // extended with the functionname
    Closure closure = new Closure(lambdaExpr.formals, lambdaExpr.body, r);
    return k.resume(closure);
  }

  /**
   * Processes cond expressions.
   */
  private Object visitCondExpr(CondExpr condExpr, Environment r, 
			       Continuation k) 
    throws InvalidTypeException, SchemeJException
  {
    try {
      for (int i = 0; i < condExpr.clauses.length; i++) {
	ClauseExpr clause = condExpr.clauses[i];
	Object testval = 
	  clause.test.eval(this, r, new PrintErrorContinuation(r, null));
	
	// if testval has boolean type
	if (testval instanceof Boolean) {
	  // if the given testval evaluates to true
	  if (((Boolean) testval).booleanValue() == true) {
	    int noexprs = condExpr.clauses[i].exprs.length;
	    Object value = null;
	    for (int j = 0; j < noexprs; j++) {
	      value = 
		clause.exprs[j].eval(this, r, 
				     new PrintErrorContinuation(r, k));
	    }
	    return k.resume(value);
	  }
	  // if the given test val evaluates to false
	  else ;
	}
	else {
	  throw new InvalidTypeException("a clause should begin with a " +
					 "expression with boolean type");
	}
      }

      // cond expr has no clause with true condition
      // HOW CAN I ADD EXCEPTION EFFECT?
      return k.resume(null);
    } catch (SchemeJException e) {
      return k.resume(e);
    }


  }

  /**
   * Processes case expressions.
   */
  private Object visitCaseExpr(CaseExpr caseExpr, Environment r, 
			       Continuation k)
    throws InvalidTypeException, SchemeJException
  {
    try {
      Object keyvalue = caseExpr.key.eval(this, r, k);
      
      for (int i = 0; i < caseExpr.clauses.length; i++) {
	//Object testval = caseExpr.clauses[i].key.accept(this);
	
	// if test value and key value has the same type
	//if (testval.getType() == keyvalue.getType) {
	if (true) {
	  // if test value and key value matches
	  // if (testval.getValue().equals(keyvalues.getValue())) {
	  if (true) {
	    //return caseExpr.clauses[i].expr.accept(this);
	    return null;
	  }
	  // need to replace the following else-stmt
	  else {
	    throw new SchemeJException();
	  }
	}
	else {
	  throw new InvalidTypeException();
	}
      }
      throw new SchemeJException();
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  }

  /**
   * Processes let expressions. For this, a new environment
   * is created and extended with the declarations wherewithin
   * the body is evaluated.
   */
  private Object visitLetExpr(LetExpr letExpr, Environment r, Continuation k) 
    throws SchemeJException 
  {
    Environment letenv = new Environment(r);

    try {
      // if the let expression is not named
      Continuation errork = new PrintErrorContinuation(r, null);
      if (!letExpr.isNamedLet()) {
	Object[] values = new Object[letExpr.decls.length];
	for (int i = 0; i < letExpr.decls.length; i++) {
	  letenv.extend(letExpr.decls[i].var.name, 
			letExpr.decls[i].init.eval(this, letenv, errork));
	}
	// the body is evaluated w.r.t the baseEnv extended with
	// the bindings
	return k.resume(letExpr.body.eval(this, letenv, errork));
					  
      }
      // else if the let expression is named
      // NOT IMPLEMENTED
      else {
	return k.resume(null);
      }
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  }

  /**
   * Processes let* expressions. 
   */
  private Object visitLetStarExpr(LetStarExpr letsExpr, Environment r,
				  Continuation k) 
    throws SchemeJException
  {
    Environment letenv = new Environment(r);

    try {
      // if the let expression is not named
      Continuation errork = new PrintErrorContinuation(r, null);
      if (!letsExpr.isNamedLet()) {
	Object[] values = new Object[letsExpr.decls.length];
	for (int i = 0; i < letsExpr.decls.length; i++) {
	  letenv.extend(letsExpr.decls[i].var.name, 
			letsExpr.decls[i].init.eval(this, letenv, errork));
	}

	// the body is evaluated w.r.t the baseEnv extended with
	// the bindings
	return k.resume(letsExpr.body.eval(this, letenv, errork));
      }
      // else if the let expression is named
      // NOT IMPLEMENTED
      else {
	return k.resume(null);
      }
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  }

  /**
   * Processes letrec expressions.
   */
  private Object visitLetrecExpr(LetrecExpr letrecExpr, Environment r,
				 Continuation k)
    throws SchemeJException
  {
    Environment letenv = new Environment(r);

    try {
      Continuation errork = new PrintErrorContinuation(r, null);
      Object[] values = new Object[letrecExpr.decls.length];

      // first extend the letenv with decl var names 
      // with some initial values, say "*"
      for (int i = 0; i < letrecExpr.decls.length; i++) {
	letenv.extend(letrecExpr.decls[i].var.name, (String) "*");
      }

      for (int i = 0; i < letrecExpr.decls.length; i++) {
	letenv.destructiveExtend(letrecExpr.decls[i].var.name, 
				 letrecExpr.decls[i].init.eval(this, letenv, 
							       errork));
      }

      // the body is evaluated w.r.t the baseEnv extended with
      // the bindings
      return k.resume(letrecExpr.body.eval(this, letenv, errork));
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  }

  /**
   * Processes sequencing expressions.
   */
  private Object visitBeginExpr(BeginExpr beginExpr, Environment r,
				Continuation k) 
    throws SchemeJException
  {
    BeginContinuation begink = 
      new BeginContinuation(beginExpr.exprs, r, k);
      
    return begink.resume(null);
  }

  /**
   * Processes do iterations.
   */
  private Object visitDoExpr(DoExpr doExpr, Environment r, Continuation k) 
    throws SchemeJException 
  {
    // NOT IMPLEMENTED
    return null;
  }

  /**
   * Processes delay expressions.
   */
  private Object visitDelayExpr(DelayExpr delayExpr, Environment r,
				Continuation k) 
    throws SchemeJException
  {
    // NOT IMPLEMENTED
    return null;
  }

  /**
   * Processes quasiquote expressions.
   */
  private Object visitQuasyQuoteExpr(QuasyQuoteExpr qqExpr, Environment r,
				     Continuation k) 
  {
    // NOT IMPLEMENTED
    return null;
  }

  /**
   * Processes define form.
   */
  private Object visitDefineForm(DefineForm defineForm, Environment r,
				 Continuation k) 
    throws SchemeJException
  {
    try {
      if (defineForm.defn instanceof LambdaExpr) {
	// for recursive function definition
	r.extend(defineForm.var.name, (String) "*");
      }
      Object value = defineForm.defn.eval(this, r,
					  new PrintErrorContinuation(r, null));
      if (value instanceof Procedure) {
	// for recursive function definition
	r.destructiveExtend(defineForm.var.name, value);
      }
      else if (value instanceof SchemeJException) {
	// unreachable // delete later
	return k.resume(value);
      }
      else {
	r.extend(defineForm.var.name, value);
      }
      
      //return k.resume(value);
      return k.resume(defineForm.var);
    } catch (SchemeJException e) {
      return k.resume(e);
    }
  }
}
