/* -*---------------------------------------------------------
 * Uses: Classic VM (build JDK-1.2-V, green threads, sunwjit)
 * -*---------------------------------------------------------
 * @(#) $Id$
 *
 * SExprParser.java:
 *   - created by Cheoljoo Jeong
 *
 * Copyleft(c) 2000. Cheoljoo Jeong. All lefts reserved.
 *
 * $Log$
 */

package kudos.schemej;

import kudos.schemej.util.*;
import kudos.lang.*;
import kudos.exns.*;
import java.util.*;

/**
 * SExprParser class accepts an s-expression and returns an
 * Expr object, which is essentially in a abstract syntax 
 * tree form.
 * Note that the parse() method of SExprParser class recurs 
 * over the recursive `data type' of SExpr.
 *
 * @author Cheoljoo Jeong
 */

public class SExprParser implements SExprVisitorI {

  /**
   * Parses an atomic expression into an Expr object.
   */
  public Object visitAtom(Atom atom) {
    switch (atom.getId()) {
      case sym.SYMBOL:
	String symbol = (String) atom.getValue();
	return new VarRefExpr(symbol);
      case sym.NUMBER:
	return new LiteralExpr((Double) atom.getValue());
      case sym.STRING:
	String str = (String) atom.getValue();
	return new LiteralExpr(str.substring(1, str.length() - 1));
      case sym.CHAR:
	return new LiteralExpr((String) atom.getValue());
      default:
	return new LiteralExpr("*mystery*");
    }
  }
  
  /**
   * Parses a list into an Expr object.
   */
  public Object visitPair(Pair pair)
    throws SExprParseException, InvalidDefineFormException,
	   WrongNumberOfRandsException 
  {
    Expr expr = null;

    int length = pair.length();
    SExpr car = pair.car;
    SExpr cdr = pair.cdr;

    try {
      if (car instanceof Atom) {
	Atom caratom = (Atom) car;
	if ((caratom.getId() == sym.SYMBOL) || 
	    (caratom.getId() == sym.QUOTE)) {
	  String keyword = (String) ((Atom) car).getValue();

	/* quote expression */
	  if (keyword.equals("quote")) {
	    return genQuoteExpr(pair);
	  }
	  /* define form */
	  else if (keyword.equals("define")) {
	    if (!(pair.cdr instanceof NullList)) {
	      if (((Pair) pair.cdr).car.isPair()) {
		return genDefineProcedureForm(pair);
	      }
	      else {
		return genDefineVariableForm(pair);
	      }
	    }
	    else {
	      throw new SExprParseException(SchemeJException.EvalError, 
					    "Huh?");
	    }
	  }
	  /* lambda expression */
	  else if (keyword.equals("lambda")) {
	    return genLambdaExpr(pair);
	  }
	  /* if expression */
	  else if (keyword.equals("if")) {
	    return genIfExpr(pair);
	  }
	  /* set! expression */
	  else if (keyword.equals("set!")) {
	    return genSetExpr(pair);
	  }
	  /* let expression */
	  else if (keyword.equals("let")) {
	    return genLetExpr(pair);
	  }
	  /* let* expressions */
	  else if (keyword.equals("let*")) {
	    return genLetStarExpr(pair);
	  }
	  /* letrec expressions */
	  else if (keyword.equals("letrec")) {
	    return genLetrecExpr(pair);
	  }
	  /* cond expressions */
	  else if (keyword.equals("cond")) {
	    return genCondExpr(pair);
	  }
	  /* case expressions */
	  else if (keyword.equals("case")) {
	    return genCaseExpr(pair);
	  }
	  /* begin expressions */
	  else if (keyword.equals("begin")) {
	    return genBeginExpr(pair);
	  }
	  /* combinations */
	  else { 
	    return genApplyExpr(pair);
	  }
	}
	// if car is not a symbol
	else {
	  throw new SExprParseException(SchemeJException.EvalError,
					"*mystery*");
	}
      }
      /* yet another combination; ((lambda (x) body) args ...) */
      else {
	return genApplyExpr(pair);
      }
    } catch (SExprParseException e) {
      throw e;
    }
  }

  /**
   * Quoted expressions.
   * Syntax: (quote <expr>)
   */
  private QuoteExpr genQuoteExpr(Pair pair) throws SExprParseException {
    String syntax = "(quote <expr>)";
    // We regard that the semantics of a quoted expression
    // is the s-expression, a pre-syntactic entity.
    if (pair.length() != 2) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }

    if (((Pair) pair.cdr).car != null) {
      return new QuoteExpr(((Pair) pair.cdr).car);
    }
    else {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  /**
   * Define forms (function definition).
   * Syntax: (define (<func> <param_1> <param_2> ...) <body>)
   */
  private DefineForm genDefineProcedureForm(Pair pair) 
    throws SExprParseException 
  {
    String syntax = "(define (<func> <param_1> <param_2> ...) <body>)";
    if (pair.length() != 3) {
      throw new SExprParseException(SchemeJException.EvalError, syntax +
				    "\nshould be length 3");
    }
    if (!((Pair) ((Pair) pair.cdr).car).car.isAtom()) {
      throw new SExprParseException(SchemeJException.EvalError, syntax+
				    "\n<func> should be a symbol");
    }

    SExpr func = ((Pair) ((Pair) pair.cdr).car).car;
    SExpr formals = ((Pair) ((Pair) pair.cdr).car).cdr;
    SExpr body = ((Pair) ((Pair) pair.cdr).cdr).car;
    String funcname;

    if (((Atom) func).getId() != sym.SYMBOL) {
      throw new SExprParseException(SchemeJException.EvalError, 
				    syntax + "\nvar should be a symbol");
    }
    else {
      // funcname is needed for checking recursive procedure defintion
      funcname = (String) ((Atom) func).getValue();
    }
    
    try {
      VarRefExpr varref = new VarRefExpr((String) ((Atom) func).getValue());
      String[] formalNames;
      if (formals instanceof NullList) {
	formalNames = new String[0];
      }
      else {
	formalNames = parseFormals((Pair) formals);
      }
      Expr bodyExpr = (Expr) body.accept(this);

      return new DefineForm(varref, 
			    new LambdaExpr(funcname, formalNames, bodyExpr));
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  /**
   * Define forms (variable definition).
   * (define <var> <defn>)
   */
  private DefineForm genDefineVariableForm(Pair pair) 
    throws SExprParseException 
  {
    String syntax = "(define <var> <defn>)";
    if (pair.length() != 3) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
    
    SExpr var = (Atom) ((Pair) pair.cdr).car;
    SExpr defn = ((Pair) ((Pair) pair.cdr).cdr).car;

    if (!var.isAtom()) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
    if (((Atom) var).getId() != sym.SYMBOL) {
      throw new SExprParseException(SchemeJException.EvalError, 
				    syntax + ": var should be a symbol");
    }

    try {
      VarRefExpr varref = new VarRefExpr((String) ((Atom) var).getValue());
      Expr expr = (Expr) defn.accept(this);
      if (expr instanceof LambdaExpr) {
	((LambdaExpr) expr).funcname = varref.name;
      }

      return new DefineForm(varref, expr);
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  /**
   * Lambda expressions.
   * Syntax: (lambda <formals> <body>)
   *         where <formals> is (<param_1> <param_2> ...)
   */
  private LambdaExpr genLambdaExpr(Pair pair) throws SExprParseException {
    String syntax = 
      "(lambda <formals> <body>)\n" + 
      " where <formals> is (<param_1> <param_2> ...)";

    if (pair.length() != 3) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }

    SExpr formals = ((Pair) pair.cdr).car;
    SExpr body = ((Pair) ((Pair) pair.cdr).cdr).car;

    try {
      String[] formalNames;
      if (!(formals instanceof NullList)) { 
	formalNames = parseFormals((Pair) formals);
      }
      else {
	formalNames = new String[0];
      }
      
      Expr bodyExpr = (Expr) body.accept(this);
      
      return new LambdaExpr(formalNames, bodyExpr);
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }
  
  /**
   * If expressions.
   * Syntax: (if <condexpr> <thenexpr> <elseexpr>)
   */
  private IfExpr genIfExpr(Pair pair) throws SExprParseException {
    String syntax =
      "(if <condexpr> <thenexpr> <elseexpr>)";

    if ((pair.length() != 4) && (pair.length() != 3)) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }

    SExpr condexpr = ((Pair) pair.cdr).car;
    SExpr thenexpr = ((Pair) ((Pair) pair.cdr).cdr).car;
    SExpr elseexpr = ((Pair) ((Pair) ((Pair) pair.cdr).cdr).cdr).car;
    try {
      Expr condExpr = (Expr) condexpr.accept(this);
      Expr thenExpr = (Expr) thenexpr.accept(this);
      if (pair.length() == 4) {
	Expr elseExpr = (Expr) elseexpr.accept(this);
	return new IfExpr(condExpr, thenExpr, elseExpr);
      }
      else {
	return new IfExpr(condExpr, thenExpr);
      }
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  /**
   * Set! expressions.
   * Syntax: (set! <var> <expr>)
   */
  private SetExpr genSetExpr(Pair pair) throws SExprParseException {
    String syntax = "(set! <var> <expr>)";

    if (pair.length() != 3) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }

    SExpr var = ((Pair) pair.cdr).car;
    SExpr val = ((Pair) ((Pair) pair.cdr).cdr).car;

    if (!var.isAtom()) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
    if (((Atom) var).getId() != sym.SYMBOL) {
      throw new SExprParseException(SchemeJException.EvalError, 
				    syntax + ": var should be a symbol");
    }

    try {
      VarRefExpr varref = new VarRefExpr((String) ((Atom) var).getValue());
      Expr valExpr = (Expr) val.accept(this);

      return new SetExpr(varref, valExpr);

    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  
  /**
   * Let expressions.
   * Syntax: (let (<binding_1> <binding_2> ...) <body>)
   *         where <binding_i> is (<var> <init>)
   *         or
   *         (let <name> (<binding_1> <binding_2> ...) <body>)
   */
  private LetExpr genLetExpr(Pair pair) throws SExprParseException {
    String name;
    SExpr bindings;
    SExpr body;

    String syntax = 
      "(let (<binding_1> <binding_2> ...) <body>)\n" +
      " where <binding_i> is (<var> <init>)" +
      " or \n" + 
      " (let <name> (<binding_1> <binding_2> ...) <body>)";

    if ((pair.length() != 3) && (pair.length() != 4)) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }

    // for anonymous let expressions
    if (pair.length() == 3) {
      bindings = ((Pair) pair.cdr).car;
      body = ((Pair) ((Pair) pair.cdr).cdr).car;
    }
    // for named let expressions
    else {
      bindings = ((Pair) ((Pair) pair.cdr).cdr).car;
      body = ((Pair) ((Pair) ((Pair) pair.cdr).cdr).cdr).car;
    }
    
    try {
      BindingExpr[] bindingExprs = parseBindings((Pair) bindings);
      Expr bodyExpr = (Expr) body.accept(this);

      if (pair.length() == 3) {
	return new LetExpr(bindingExprs, bodyExpr);
      }
      else {
	if (pair.car.isAtom()) {
	  name = (String) ((Atom) pair.car).getValue();
	  return new LetExpr(name, bindingExprs, bodyExpr);
	}
	else {
	  throw new SExprParseException(SchemeJException.EvalError,
					"strange named let expressions");
	}
      }
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  /**
   * Let* expressions.
   * Syntax: (let* (<binding_1> <binding_2> ...) <body>)
   *         where <binding_i> is (<var> <init>)
   */
  private LetStarExpr genLetStarExpr(Pair pair) throws SExprParseException {
    String syntax = 
      "(letstar (<binding_1> <binding_2> ...) <body>)\n" +
      " where <binding_i> is (<var> <init>)";

    if (pair.length() != 3) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
    
    SExpr bindings = ((Pair) pair.cdr).car;
    SExpr body = ((Pair) ((Pair) pair.cdr).cdr).car;
    
    try {
      BindingExpr[] bindingExprs = parseBindings((Pair) bindings);
      Expr bodyExpr = (Expr) body.accept(this);
      return new LetStarExpr(bindingExprs, bodyExpr);
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  /**
   * Letrec expressions.
   * Syntax: (letrec (<binding_1> <binding_2> ...) <body>)
   *         where <binding_i> is (<var> <init>)
   *         or
   *         (let <name> (<binding_1> <binding_2> ...) <body>)
   */
  private LetrecExpr genLetrecExpr(Pair pair) throws SExprParseException {
    String syntax = 
      "(letrec (<binding_1> <binding_2> ...) <body>)\n" +
      " where <binding_i> is (<var> <init>)";

    if (pair.length() != 3) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
    
    SExpr bindings = ((Pair) pair.cdr).car;
    SExpr body = ((Pair) ((Pair) pair.cdr).cdr).car;
    
    try {
      BindingExpr[] bindingExprs = parseBindings((Pair) bindings);
      Expr bodyExpr = (Expr) body.accept(this);
      return new LetrecExpr(bindingExprs, bodyExpr);
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  /**
   * Cond expressions.
   * Syntax: (cond <clause_1> <clause_2> ...)
   *         where each <clause_i> is (<test> <expression>)
   */
  private CondExpr genCondExpr(Pair pair) throws SExprParseException {
    String syntax = 
      "(cond <clause_1> <clause_2> ...)\n" +
      " where each <clause_i> is (<test> <expression>)";

    if (pair.length() <= 1) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }

    try {
      ClauseExpr[] clauses = parseClauses((Pair) pair.cdr);
      return new CondExpr(clauses);
    } catch (SExprParseException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  /** 
   * Case expressions.
   * Syntax: (case <key> <clause_1> <clause_2> ...)
   *         where <key> is any expr and
   *         <clause_1> is ((<datum_1> ...) <expr_1> ...)
   */
  private CaseExpr genCaseExpr(Pair pair) throws SExprParseException {
    String syntax = 
      "(case <key> <clause_1> <clause_2> ...)\n" +
      " where <key> is any expr and <clause_1> is" + 
      " ((<datum_1> ...) <expr_1> ...)";

    if (pair.length() < 3) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }

    try {
      Expr key = (Expr) pair.car.accept(this);
      CaseClauseExpr[] caseClauses = parseCaseClauses((Pair) pair.cdr);
      return new CaseExpr(key, caseClauses);
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }

  /**
   * Begin expressions.
   * Syntax: (begin <rand_1> <rand_2> ...)
   *	     where <rator> is an expression
   */
  private BeginExpr genBeginExpr(Pair pair) throws SExprParseException {
    String syntax = 
      "(begin <rand_1> <rand_2> ...)\n" +
      " where <rator> is an expression";
    if (pair.length() < 2) {
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
    
    try {
      Expr[] exprs = parseExprs((Pair) pair.cdr);
      return new BeginExpr(exprs);
    } catch (SExprParseException e) {
      throw e;
    }
  }

  /**
   * Combinations.
   * Syntax:(<rator> <rand_1> <rand_2> ...)
   *        where <rator> is procedure name only!
   */
  private ApplyExpr genApplyExpr(Pair pair) throws SExprParseException {
    String syntax =
      "(<rator> <rand_1> <rand_2> ...)\n";
    try {
      Expr rator = (Expr) pair.car.accept(this);
      Expr[] rands;
      if (!(pair.cdr instanceof NullList)) {
	rands = parseExprs((Pair) pair.cdr);
      }
      else {
	rands = new Expr[0];
      }

      return new ApplyExpr(rator, rands);
    } catch (SchemeJException e) {
      System.out.println(e.getMessage());
      throw new SExprParseException(SchemeJException.EvalError, syntax);
    }
  }


  /**
   * Parses a list of parameters into an array of String's.
   */
  private String[] parseFormals(Pair pair) throws SExprParseException {
    String[] formals = new String[pair.length()];
    
    int formalcount = 0;
    while (pair != null) {
      if (pair.car.isAtom()) {
	formals[formalcount] = (String) ((Atom) pair.car).getValue();
      } 
      else {
	throw new SExprParseException(SchemeJException.EvalError,
				      "parseFormals error");
      }
      if (pair.cdr instanceof NullList) {
	pair = null;
      }
      else {
	pair = (Pair) pair.cdr;
      }
      formalcount++;
    }
    return formals;
  }

  /**
   * Parses a list of bindings into an array of BindingExpr's
   * of form "((<var> <init>) (<var> <init>) ...)".
   */
  private BindingExpr[] parseBindings(Pair pair) throws SExprParseException {
    BindingExpr[] bindings;
    BindingExpr binding;

    if (pair.car.isPair()) {
      binding = parseBinding((Pair) pair.car);
    } 
    else {
      throw new SExprParseException(SchemeJException.EvalError,
				    "binding should be in a list form");
    }
    
    try {
      // if not last
      if (!(pair.cdr instanceof NullList)) {
	BindingExpr[] temp = parseBindings((Pair) pair.cdr);
	bindings = new BindingExpr[temp.length + 1];
	bindings[0] = binding;
	for (int i = 0; i < temp.length; i++) {
	  bindings[i+1] = temp[i];
	}
      }
      // if last
      else {
	bindings = new BindingExpr[1];
	bindings[0] = binding;
      }
      return bindings;
    } catch (SExprParseException e) {
      throw e;
    }
  }

  /**
   * Parse a single binding into a BindingExpr object.
   */
  private BindingExpr parseBinding(Pair pair) throws SExprParseException {
    VarRefExpr varref;
    Expr init;

    if (pair.length() != 2) {
      throw new SExprParseException(SchemeJException.EvalError,
				    "binding should have length 2");
    }

    if (pair.car.isAtom()) {
      varref = new VarRefExpr((String) ((Atom) pair.car).getValue());
    }
    else {
      throw new SExprParseException(SchemeJException.EvalError,
				    "a binding should begin with a symbol");
    }

    try {
      init = (Expr) ((Pair) pair.cdr).car.accept(this);
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError,
				    "binding parse error");
    }

    return new BindingExpr(varref, init);
  }

  /**
   * Parses a list of clauses into an array of ClauseExpr's.
   */
  private ClauseExpr[] parseClauses(Pair pair) throws SExprParseException
  {
    ClauseExpr[] clauses = new ClauseExpr[pair.length()];

    try {
      int clauseCount = 0;
      while (pair != null) {
	ClauseExpr clauseExpr = parseClause((Pair) pair.car);
	clauses[clauseCount] = clauseExpr;
	
	if (pair.cdr instanceof NullList) {
	  pair = null;
	}
	else {
	  pair = (Pair) pair.cdr;
	}
	clauseCount++;
      }
      return clauses;
    } catch (SExprParseException e) {
      throw e;
    }    
    
  }

  /**
   * Parses a single clause into a ClauseExpr object.
   */
  private ClauseExpr parseClause(Pair pair) throws SExprParseException {
    if (pair.length() != 2) {
      throw new SExprParseException(SchemeJException.EvalError, 
				    "clause parse error");
    }

    try {
      Expr test = (Expr) pair.car.accept(this);
      Expr[] exprs = parseExprs((Pair) pair.cdr);
      return new ClauseExpr(test, exprs);
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, 
				    "clause parse error");
    }
  }

  /**
   * Parses a list of expressions into a list of Expr objects.
   */
  private Expr[] parseExprs(Pair pair) throws SExprParseException {
    Expr[] exprs = new Expr[pair.length()];

    try {
      int exprcount = 0;
      while (pair != null) {
	exprs[exprcount] = (Expr) pair.car.accept(this);
	if (pair.cdr instanceof NullList) {
	  pair = null;
	}
	else {
	  pair = (Pair) pair.cdr;
	}
	exprcount++;
      }
      return exprs;
    } catch (SchemeJException e) {
      throw new SExprParseException(SchemeJException.EvalError, 
				    "parseExprs error");
    }
  }

  /**
   * Parses a list of case clauses into an array of CaseClauseExpr's.
   */
  private CaseClauseExpr[] parseCaseClauses(Pair pair) 
    throws SExprParseException 
  {
    CaseClauseExpr[] clauses = new CaseClauseExpr[pair.length()];

    try {
      int clauseCount = 0;
      while (pair != null) {
	SExpr clause = pair.car;
	if (((Pair) ((Pair) clause).car).isPair()) {
	  // build data list
	  Pair datalist = (Pair) ((Pair) clause).car;
	  Atom[] data = new Atom[((Pair) clause).car.length()];
	  int dataCount = 0;
	  while (datalist != null) {
	    if (datalist.car.isAtom()) {
	      data[dataCount] = (Atom) datalist.car;
	    }
	    else {
	      throw new SExprParseException(SchemeJException.EvalError,
					    "a datum should be a symbol");
	    }
	    
	    if (datalist.cdr instanceof NullList) {
	      datalist = null;
	    }
	    else {
	      datalist = (Pair) datalist.cdr;
	    }
	    dataCount++;
	  }

	  Expr[] exprs = parseExprs((Pair) pair.cdr);
	  clauses[clauseCount] = new CaseClauseExpr(data, exprs);
	}
	else {
	  throw new SExprParseException(SchemeJException.EvalError,
					"case clause should have " +
					"a data list");
	}
	if (pair.cdr instanceof NullList) {
	  pair = null;
	}
	else {
	  pair = (Pair) pair.cdr;
	}
	clauseCount++;
      } 
      return clauses;
    } catch (SExprParseException e) {
      throw e;
    }
  }

  /**
   * Visitor method for the NullList.
   */
  public Object visitNullList(NullList nullList) {
    return null;
  }
}
